/* History: 
   ========
 VB Jun 1999: VDS SW for ITR
 VB Jul 1999: two helix versions with different pipelinelength  
              with new variables in header, FED, Chip
 VB 20/9/99 : chip->flag = 2 
 VB 24/10/99: no target 
 VB 24/10/99: APS type removed
 VB 10/12/99: some different datavalid-offsets
              in fetch_fed routines changed
              therefor check_event etc removed
 VB 10/01/00: fill height of columnnumber
 AL march 00: treatment of sparsified records adapted to the
              ITR sparse format (changes incorporated by wg)
*/ 


#include <stdio.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <ctype.h>

/* DAQ includes */
#include "etb.h"
#include "esp.h"
#include "grr.h"

#include "aframelib.h"
#include "iol.h"

#include "itrleda.h"
#include "ITR_silraw.h"



/* list of bad chips */
int ITR_badchip[DIM_NCHIP];
int ITR_nbadchip = 0;

/* variables for event checks */
int ITR_silraw_evno;
int ITR_silraw_evfail_tot;
int ITR_silraw_linkfail_tot;
int ITR_silraw_cellfail_tot_short;
int ITR_silraw_cellfail_tot_long;
int ITR_silraw_chipfail[DIM_NCHIP];
int ITR_silraw_chipid[DIM_NCHIP];

/* ================================================================

   Auxiliary routines to deal with the 1998 data 

   ================================================================ */

#define NOFFS 16   /* offset range to be checked in repair mode */
/*for the positions you should look in marvin_header*/
#define majorpositioninsparse 0     /*position of the major number in sparse data */
#define minorpositioninsparse 1     /*position of the minor number in sparse data */
#define cellshortpositioninsparse 2 /*position of cellshort in sparse data */
#define celllongpositioninsparse 3  /*position of celllong in sparse data */
#define formatpositioninsparse 4    /*position of format in sparse data */
#define nclpositioninsparse 5       /*position of ncl in sparse data */

int ITR_silraw_fit_cell(unsigned char *data) {
  return ITR_silraw_fit_cell_pipe (data, NCELLHELIXLONG);
}


int ITR_silraw_fit_cell_pipe(unsigned char *data, int pipelen) 
/*-----------------------------------------------------------------------------
 * Determine the cell number for a helix 
 * ===================================== 
 * A two-step procedure is being used to infer the pipeline cell number.
 *
 * First a simple algorithm is tried, where the bit pattern is for the 
 * cell number is extracted directly from the normalized -- mapped into
 * the range [0,1] -- digitizings. 
 * 
 * If this procedure results in an invalid column number, an unweighted 
 * least squares fit is performed. Here for a given assumed cell number 
 * the digitizings d are fitted to the model d=a+b*s, where s is the vector 
 * of zeros and ones that defines the cell number, a and offset and b the 
 * difference between a digitized zero and a digitized one. The fit 
 * hypothesis runs over all possible pipeline cells. For each hypothesis 
 * the parameters a and b are determined by a least squares fit. Since the 
 * levels are inverted, the best fit result with b<0 is returned. 
 *
 * Author: Michael Schmelling / 2-Sep-1998
 *---------------------------------------------------------------------------*/
{
  static int mask[8] = {1,2,4,8,16,32,64,128}; /* bitmap for single bits */

  int    i, imax, imin, s;
  int    ide, ids, idd, ise, ival, cell;
  float  de,  ds,  dd,  se,  det, a, b, chi2, chi2min, offset, amplitude;
/*---------------------------------------------------------------------------*/

  /*
   *  try a simple high-low analysis of the cell-number digitizings
   */ 
  
  imax = *(data);
  imin = *(data);
  for(i=1; i<8; i++) {
    ival = *(data+i);
    if(ival>imax) imax = ival;
    if(ival<imin) imin = ival;
  }

  cell = 0.; 
  if(imax-imin<10) return cell;
 
  for(i=0; i<8; i++) {
    if(((float)(*(data+i)-imin)/(float)(imax-imin)) < 0.5) cell += mask[i];
  } 
  
  if(cell<pipelen) return cell;

  /*
   *  if that fails do the more sophisticated fit 
   */

  /* sum digitizings */

  ide = 0;
  idd = 0;
  for(i=0; i<8; i++) {
    ival = *(data+i);
    ide += ival;
    idd += ival*ival;
  } 
  de = (float)ide;
  dd = (float)idd;

  /* start with hypothesis cell=0 */
  
  cell      = 0;
  chi2min   = dd - de*de/8.;
  offset    = de/8.;
  amplitude = 0;

  if(chi2min < 50.) return cell;

  /* loop over all candidate cell numbers */

  for(s=1; s<pipelen; s++) {
    ise = 0;
    ids = 0;
    for(i=0; i<8; i++) {
      if(s & mask[i] ) {
        ids += *(data+i);
        ise += 1;
      }
    } 
    se   = (float)ise;
    ds   = (float)ids;
    det  = (8.-se)*se;
    a    = se*(de-ds)   /det;     /* offset */
    b    = (8.*ds-se*de)/det;     /* amplitude */
    chi2 = dd - a*de - b*ds;  
    if(chi2<chi2min && b<0) {
      chi2min   = chi2;
      cell      = s;
      offset    = a;
      amplitude = b;
    }
  }
 
  return cell; 
}

/* ----------------------------------------------------------------
 *
 * Read list of bad chips
 *
 * List of bad chips has to be given in hexa-decimal
 *
 * ---------------------------------------------------------------- */
int ITR_silraw_read_bad_chip_list()
{
  FILE *infile;

  int   i, chip;

  char  text[256];
  int   blank;

  char mess[256];
  char *routine = "ITR_silraw_read_bad_chip_list";

  if ( !strcmp( ITR_silraw_iol.bad_chip_list.value, "none" ) ) {
    aframe_mess(routine,GRR_SINFO, "No list of bad chips given.");
    return GRR_OK;
  }

  /* open file with list of bad chips */
  infile = fopen( ITR_silraw_iol.bad_chip_list.value, "r" );
  if ( infile == NULL ) {
    aframe_mess(routine,GRR_SERROR,
		"Could not open file %s with list of bad chips",
		ITR_silraw_iol.bad_chip_list.value );
    return GRR_FISHY;
  }
  aframe_mess(routine,GRR_SINFO, "Reading list of bad chips from %s",
	      ITR_silraw_iol.bad_chip_list.value );
  
  /* read list of bad chips im hex!! */
  while( fgets( text, 256, infile ) ) {
    blank = 0; 
    while ( isspace(text[blank]) )
      blank++;
    if ( isdigit(text[blank]) != 0 ) {
      sscanf( text, "%x", &chip ); 
      ITR_badchip[ITR_nbadchip] = chip;
      ITR_nbadchip++;
    }
  }

  /* dump bad chip list */
  aframe_mess(routine,GRR_SINFO,
	      "List of bad chips (not being considered in read-out):");
  strcpy (mess, "");
  for ( i=0; i<ITR_nbadchip; i++ ) {
    sprintf(mess, "%s %4x", mess, ITR_badchip[i] );
    if ( (i+1)%12 == 0 ) {
      aframe_mess(routine,GRR_SINFO,mess);
      strcpy (mess, "");
    }
  }
  aframe_mess(routine,GRR_SINFO,mess);

  /* close bad chip file */
  fclose( infile );

  return GRR_OK;
}

/* ----------------------------------------------------------------
 *
 * Check, if chip with id 'chipid' is in bad chip list
 *
 * Returns 1 if yes, 0 if not
 *
 * ---------------------------------------------------------------- */
int ITR_silraw_bad_chip( int chipid )
{
  int   i, chip_is_bad;

  /* char *routine = "ITR_silraw_bad_chip"; */

  chip_is_bad = 0;
  for ( i=0; i<ITR_nbadchip; i++ )
    if ( ITR_badchip[i] == chipid )
      chip_is_bad = 1;

  return chip_is_bad;
}



/* ================================================================

   ROUTINES TO GET ITR_Silraw_Event_t FROM RAW DATA (EVENT, SLB OR FED FORMAT)

   ================================================================ */


/* ----------------------------------------------------------------
 *
 * Fill ITR_Silraw_Event_t from EVENT block
 *
 * Inputs : char *event : Pointer to EVENT block
 *
 * ---------------------------------------------------------------- */
int ITR_silraw_get_event( EVENT_HDR *event )
{
  int       rc;
  int       nfed, ndaq;
  EVENT_HDR imp_event;  /* event header in imported format */
  ETB_fed   etb_feds[DIM_NFED];
  int       cellshort, celllong;

  static EVT_DATA_RECORD daq;          /* imported daq event block */

  ITR_Silraw_Event_t *silraw;

  char *routine = "ITR_silraw_get_event";

  /* ---------------------------------------------------------- */

  /* fetch ITR_Silraw_Event_t */
  silraw = ITR_silraw_fetch_event();
  if ( !silraw )
    return GRR_FISHY;

  if ( ITR_silraw_iol.debug_mode.value )
    aframe_mess(routine,GRR_SINFO, "Raw data in EVENT format.");

  /* get all FEDs (=SLB sub-blocks) from ITR */
  rc = ETB_get_feds( event, COMP_ID_ITR, DIM_NFED, etb_feds, &nfed ); 
  if ( rc != ETB_ok || nfed <= 0 ) {
    if (  nfed > 0 && ITR_silraw_iol.debug_mode.value )
      ETB_dump_feds( stdout, event, etb_feds, nfed);
    aframe_mess(routine,GRR_SERROR,
		"Could not find FED blocks, nfed = %d, ETB rc = 0x%x",
		nfed, rc);
    return GRR_FISHY;
  }
  if ( ITR_silraw_iol.debug_mode.value )
    aframe_mess(routine,GRR_SINFO,
		"Found %d ITR FEDs in this event.", nfed );


  /* process list of FEDs */
  rc = ITR_silraw_fetch_feds_from_etb_list( etb_feds, nfed );
  if ( rc != GRR_OK )
      return rc;
 

  /* get DAQ event block */
  rc = ETB_get_daqs( event, 1, &daq, &ndaq );
  if ( rc != ETB_ok || ndaq <= 0 ) {
    aframe_mess(routine,GRR_SWARNING,
		"Could not find DAQ event block, rc=%d, ndaq=%d",
		rc, ndaq );
    silraw->daq = 0;
  }
  else
    silraw->daq = &daq;
 
  /* fill ITR_Silraw_header_t */
  if ( silraw->daq ) {
    silraw->header.major  = silraw->daq->data.run_number;
    silraw->header.evtime = silraw->daq->data.event_time;
  }
  ESP_get_event_hdr( event, &imp_event );
  silraw->header.sltevt = (int) imp_event.sltevt;
  silraw->header.fltevt = (int) etb_feds[0].slb->fltevt;
  silraw->header.bxid   = (int) etb_feds[0].hdr.bxid;
  
  rc = ITR_silraw_fetch_cell_from_data( &cellshort, &celllong );
  if (rc != GRR_OK )
    return rc;
  /*VB*/
  silraw->header.cellshort = (unsigned short) cellshort;
  silraw->header.celllong  = (unsigned short) celllong;
  silraw->header.flag = 0;    /*for now */


  /* dump debug output if requested */
  if ( ITR_silraw_iol.debug_mode.value )
    aframe_mess(routine,GRR_SINFO,
		"Silraw header: Cellshort = %d,long=%d, FLT-EVT = %d, BX ID = %d",
		silraw->header.cellshort, silraw->header.celllong,
		silraw->header.fltevt,silraw->header.bxid );

  return GRR_OK;
}

/* ----------------------------------------------------------------
 *
 * Fill ITR_Silraw_Event_t from list of SLB blocks
 *
 * Inputs :  SLB_HDR **slb     : Ptr to list of ptr's to slb blocks
 *           int     nslb      : number of SLB blocks in list
 *           int     ext_format: format of SLB blocks
 *
 * ---------------------------------------------------------------- */
int ITR_silraw_get_CIF_event( SLB_HDR **slb, int nslb, int ext_format )
{
  int             islb, nslbfed, rc;
  int             nfed;
  /* SLB_HDR         imp_slb; */    /* SLB header in imported format */
  ETB_fed         etb_feds[DIM_NFED];
  int             cellshort, celllong;

  ITR_Silraw_Event_t *silraw;

  char *routine = "ITR_silraw_get_CIF_event";

  /* ---------------------------------------------------------- */

   /* fetch ITR_Silraw_Event_t */
  silraw = ITR_silraw_fetch_event();
  if ( !silraw )
    return GRR_FISHY;

  if ( ITR_silraw_iol.debug_mode.value )
    aframe_mess(routine,GRR_SINFO, "Raw data in SLT format.");

  /* loop over SLBs to get all FEDs (=SLB sub-blocks) */
  nfed = 0;
  for ( islb=0; islb<nslb; islb++ ) {

    rc = ETB_get_feds_from_slb( slb[islb], ext_format, DIM_NFED-nfed,
				&(etb_feds[nfed]), &nslbfed );
    if ( rc != ETB_ok || nslbfed < 0 ) {
      aframe_mess(routine,GRR_SERROR,
		  "Could not find FED blocks, nfed=%d, islb=%d, ETB rc=%d",
		  nslbfed, islb, rc);
      return GRR_FISHY;
    }
    if ( ITR_silraw_iol.debug_mode.value )
      aframe_mess(routine,GRR_SINFO, "Found %d FED blocks in SLB %d",
		  nslbfed, islb);
    nfed += nslbfed;
  }
  if ( ITR_silraw_iol.debug_mode.value )
    aframe_mess(routine,GRR_SINFO, "=> Found %d FED blocks in total", nfed);
  
  /* process list of FEDs */
  rc = ITR_silraw_fetch_feds_from_etb_list( etb_feds, nfed );
  if ( rc != GRR_OK )
    return rc;

  /* fill ITR_Silraw_header_t */
  silraw->header.fltevt = (int) etb_feds[0].slb->fltevt;
  silraw->header.sltevt = 0;
  silraw->header.bxid   = (int) etb_feds[0].hdr.bxid;
  rc = ITR_silraw_fetch_cell_from_data( &cellshort, &celllong );
  if (rc != GRR_OK )
    return rc;
  /*VB*/
  silraw->header.cellshort = (unsigned short) cellshort;
  silraw->header.celllong  = (unsigned short) celllong;
  silraw->header.flag = 0;    /*for now */

  /* No daq event blocks known */
  silraw->daq        = 0;
     
  /* dump debug output if requested */
  if ( ITR_silraw_iol.debug_mode.value )
    aframe_mess(routine,GRR_SINFO,
		"Cellshort = %d, long=%d,  FLT-EVT = %d, BX ID = %d",
		silraw->header.cellshort, silraw->header.celllong, 
		silraw->header.fltevt,silraw->header.bxid );

  /* dump further debug output if requested */
  ITR_silraw_dump_event();

  return GRR_OK;
}

/* ----------------------------------------------------------------
 *
 * Fill ITR_Silraw_Event_t from SLB event
 *
 * Inputs : char *slb : Pointer to SLB block
 *
 * ---------------------------------------------------------------- */
int ITR_silraw_get_slb( SLB_HDR *slb )
{
  int        rc;
  int        nfed;
  ETB_fed    etb_feds[DIM_NFED];
  int        cellshort, celllong;

  ITR_Silraw_Event_t *silraw;

  char *routine = "silraw_get_slb";

  /* ---------------------------------------------------------- */

   /* fetch ITR_Silraw_Event_t */
  silraw = ITR_silraw_fetch_event();
  if ( !silraw )
    return GRR_FISHY;

  if ( ITR_silraw_iol.debug_mode.value )
    aframe_mess(routine,GRR_SINFO, "Raw data in SLB format.");
    
  /* get all FEDs (=SLB sub-blocks) from SLB block */
  rc = ITR_silraw_etb_get_feds_from_slb( slb, DIM_NFED, etb_feds, &nfed );
  if ( rc != ETB_ok || nfed <= 0 ) {
    aframe_mess(routine,GRR_SERROR, "Could not find FED blocks, nfed = %d.", nfed);
    return GRR_FISHY;
  }      

  /* process list of FEDs */
  rc = ITR_silraw_fetch_feds_from_etb_list( etb_feds, nfed );
  if ( rc != GRR_OK )
    return rc;

  /* fill ITR_Silraw_header_t */
  silraw->header.fltevt = (int) etb_feds[0].slb->fltevt;
  silraw->header.sltevt = 0;
  silraw->header.bxid   = (int) etb_feds[0].hdr.bxid;
  rc = ITR_silraw_fetch_cell_from_data( &cellshort, &celllong );
  if (rc != GRR_OK )
    return rc;
  /*VB*/
  silraw->header.cellshort = (unsigned short) cellshort;
  silraw->header.celllong  = (unsigned short) celllong;
  silraw->header.flag = 0;    /*for now */

  /* No daq event blocks known */
  silraw->daq        = 0;
    
  /* dump debug output if requested */
  if ( ITR_silraw_iol.debug_mode.value )
    aframe_mess(routine,GRR_SINFO,
		"Cellshort = %d, long=%d, FLT-EVT = %d, BX ID = %d",
		silraw->header.cellshort,silraw->header.celllong, 
		silraw->header.fltevt, silraw->header.bxid );

  return GRR_OK;
}

#define MX_CHIP_CHAN 128
#define MX_FED_CHAN  288

/* ----------------------------------------------------------------
 *
 * Expand sparsified data in etb_fed and update
 * etb_fed header
 *
 * Inputs    : ETB_FED *etb_fed : Pointer to ETB_fed
 *
 * author    : Arnd Lange
 * last issue: 31.03.2000
 * version   : 1.0a
 *             to be changed: MX_FED_CHAN    # max. channels of FED; at now as #define in routine
 *                            MX_CHIP_CHAN   # max. channels per chip; at now as #define in routine
 * ---------------------------------------------------------------- */
int expand_etb_fed( ETB_fed *etb_fed)
{ int   i, j, k, n,
        rc,
        ptr = 2, 
        offs,
        offs_0, offs_1,
        firstch[2],
        average[2],
        first_ch_offs = -1,
        datasize = MX_FED_CHAN;
  short superlayer;
  char  *data,
        fed_data[2][MX_CHIP_CHAN];
  char  *routine = "expand_etb_fed";
  int   fedid;

  static ETB_fed   *etbfed[0x4FFF] = { NULL };


  fedid = etb_fed->hdr.fedid;

  if(etbfed[fedid] == NULL) 
  { etbfed[fedid]       = malloc(sizeof(ETB_fed));
    etbfed[fedid]->slb  = malloc(sizeof(SLB_HDR));
    etbfed[fedid]->data = malloc(datasize);
  }

  if ((superlayer = ITR_silraw_get_superlayer(etb_fed->hdr.fedid)) == 0) printf("%s: cant find superlayer\n",routine);
  else  first_ch_offs = ITR_silraw_get_datavalidoffset(superlayer);

  /* --wg-- printf("%s: first_ch_offs %d\n",routine,first_ch_offs); */

  /*  first_ch_offs = 6;*/
  firstch[0] = first_ch_offs;                       /* first data channel for 1. chip */
  firstch[1] = firstch[0] + MX_CHIP_CHAN + 8;       /* first data channel for 2. chip */


  if (first_ch_offs == -1) { printf("%s: cant find offset\n",routine); return GRR_FISHY;}

  /* initialize data array */

  /* keep hdr.fedid, header, headersize
  // change length etb_fed->hdr.length  -> end of of routine */
  data       = (char*) etb_fed->data;

  if(data[0] < 0)  average[0] = -1;     else { average[0] = data[ptr]; ptr++; }
  if(data[1] < 0)  average[1] = -1;     else { average[1] = data[ptr]; ptr++; }

  /* test output of sparsified FED */
  /* --wg--
     printf("------- etb FED -------\n");
     printf("%s: FED  ID 0x%04x RecId 0x%04x   length: %d\n",routine,
     etb_fed->hdr.fedid,etb_fed->hdr.recid,etb_fed->hdr.length);

     for(i=0;i<etb_fed->hdr.length - sizeof(SLB_SUBHDR);i++)
     { if(i%16 == 0) printf("\n%4d: ",i);
     printf("%4d ",data[i]);
     }
  */

  for(k=0;k<=1;k++)
  { for(i=0;i<MX_CHIP_CHAN;i++)    fed_data[k][i] = 0;      /* set data to 0 */

    if(data[k] < 0)                                    /* no valid data in k. chip of FED */
      ;                                                   /* nothing to be done */
    else
    { for(i=0;i<MX_CHIP_CHAN;i++)    fed_data[k][i] = average[k];            /* set average value to all channels */

      offs_0 = ptr;
      offs_1 = data[k] + offs_0;

      for(j=0;j<data[k];j++)
	/*while(ptr < offs_1);*/
      { offs = data[ptr];
        if(offs < 0) { offs = -offs;
	    	       n = 1;
                       ptr++;
                     }
        else         { n = data[ptr + 1];
                       ptr += 2;
                     }

        for(i=0;i<n;i++)  fed_data[k][i+offs] = data[i+ptr]; 

        ptr += n;
        if(ptr >= offs_1) break;
      }
    }
  }

  /* expand data part of etb_fed */
  etbfed[fedid]->hdr.length = datasize + sizeof(SLB_SUBHDR);
  etbfed[fedid]->hdr.fedid  = etb_fed->hdr.fedid;
  etbfed[fedid]->hdr.recid  = etb_fed->hdr.recid;

  data       = (char*) (etbfed[fedid]->data);
  for(i=0;i<firstch[0];i++)  data[i] = 100; 
  for(i=firstch[1]-8;i<firstch[1];i++)  data[i] = 100;
  for(i=firstch[1]+MX_CHIP_CHAN;i<MX_FED_CHAN;i++)  data[i] = 100;

  for(k=0;k<=1;k++)
    for(i=0;i<MX_CHIP_CHAN;i++)    data[i+firstch[k]] = fed_data[k][i];
    
  /* test output of expanded FED */
  /* --wg--
     printf("\n------------- expanded etb FED ---------\n");fflush(stdout);
     printf("%s: FED  ID 0x%04x RecId 0x%04x   length: %d, pointer: %p\n",routine,
     etbfed->hdr.fedid,etbfed->hdr.recid,etbfed->hdr.length,etbfed);fflush(stdout);
     
     for(i=0;i<MX_FED_CHAN;i++)
     { if(i%16 == 0) printf("\n%4d: ",i);
     printf("%4d ", *(((char* ) (etbfed->data))+i)); fflush(stdout);
     }
     printf("\n");fflush(stdout);
  */

  if ( (rc = ITR_silraw_fetch_fed_from_etbfed( etbfed[fedid] )) != GRR_OK ) 
  {  aframe_mess(routine,GRR_SERROR, "Could not read FED.");
     return GRR_SERROR;
  }

  return GRR_OK;
}

/* ----------------------------------------------------------------
 *
 * Fetch SLB sub blocks from list of ETB_fed's.
 * The list may contain raw and/or sparsified data blocks.
 *
 * Inputs : ETB_FED *etb_fed : Pointer to first ETB_fed in list
 *          int     nfed     : Number of ETB_fed's in list
 *
 * ---------------------------------------------------------------- */
int ITR_silraw_fetch_feds_from_etb_list( ETB_fed *etb_fed, int nfed )
{
  int        ifed, rc;
  SLB_SUBHDR imp_slbsub;  /* SLB sub header in imported format */
  int        error;

  /* variables and arrays for faulty and missing feds */
  int   i, j, found;
  static int fed_tot[DIM_NFED];
  int   fed_ok[DIM_NFED];
  int   fed_faulty[DIM_NFED];
  int   fed_missing[DIM_NFED];
  int   nok, nfaulty, nmissing;
  static int event = 0;   /* event count          */
  static int ntot  = 0;   /* total number of FEDs */

  ITR_Silraw_Event_t *silraw;

  char mess[2048];  /*VB this string has to be big enough! */
  char *routine = "ITR_silraw_fetch_feds_from_etb_list";

  /* ---------------------------------------------------------- */

  /* fetch ITR_Silraw_Event_t */
  silraw = ITR_silraw_fetch_event();
  if ( !silraw )
    return GRR_FISHY;

  /* set counts to 0 */
  silraw->nfed    = 0;
  silraw->nchip   = 0;
  silraw->nsparse = 0;

  /* loop over FEDs (well, SLB sub blocks) */
  error = 0;
  for ( ifed=0; ifed<nfed; ifed++ ) {

    /* import SLB sub header to local format */
    imp_slbsub = etb_fed[ifed].hdr;

    if ( ITR_silraw_iol.debug_mode.value )
      aframe_mess(routine,GRR_SINFO,
		  "Reading %d.FED: Length=%d, FEDid=0x%x, recid=0x%x, bxid=%d",
		  ifed, imp_slbsub.length, imp_slbsub.fedid, 
		  imp_slbsub.recid, imp_slbsub.bxid );

    /* check for sparsified data first */
    if (  ( ITR_SILRAW_FED_SPARSE(imp_slbsub.recid)       && 
	    ITR_silraw_iol.get_faulty_feds.value != 1   ) ||  
	  ( ITR_SILRAW_FED_SPARSE_ERROR(imp_slbsub.recid) && 
	    ITR_silraw_iol.get_faulty_feds.value != 0 )       )   { 

      if ((unsigned short) imp_slbsub.length > 0x7FFF) {
	aframe_mess(routine, GRR_SWARNING,
		    "Sparsified data block too large: %d bytes", 
		    (unsigned short) imp_slbsub.length);
	return GRR_FISHY;
      }
	
      silraw->sparse[silraw->nsparse].length = (unsigned short) imp_slbsub.length
                                                - sizeof(SLB_SUBHDR);
      silraw->sparse[silraw->nsparse].recid  = imp_slbsub.recid;
      silraw->sparse[silraw->nsparse].data   = etb_fed[ifed].data;

      silraw->nsparse++;
      if ( silraw->nsparse > DIM_NSPARSE ) {
	aframe_mess(routine,GRR_SWARNING,
		    "Too many sparsified FED blocks (> %d)", DIM_NSPARSE );
	silraw->nsparse--;
      }
    }

    /*    now check for normal raw data (without error)
     * OR only faulty data, if requested with iol       */
    else if ( ( ITR_SILRAW_FED_RAW(imp_slbsub.recid)
		&& ITR_silraw_iol.get_faulty_feds.value != 1 )
	     || ( ITR_SILRAW_FED_RAW_ERROR(imp_slbsub.recid)
		&& ITR_silraw_iol.get_faulty_feds.value != 0 ) ) {

      rc = ITR_silraw_fetch_fed_from_etbfed( &etb_fed[ifed] );
      if ( rc != GRR_OK ) {
	aframe_mess(routine,GRR_SERROR, "Could not read %d-th FED.", ifed);
	error++;
      }      
    }
  }
  if ( error )
    return GRR_FISHY;

  /* get list of all FEDs expected from itrleda */
 
  if ( ntot == 0 && ITR_silraw_iol.fed_statistics_period.value > 0 ) {
    LedaTableIterator( ITR_Fed, fedIt );
    for( fedIt.begin(); fedIt.valid(); fedIt.next() ) {
      LedaTableRow(ITR_Fed) *fed  = fedIt.get();
      LedaRelationIterator(fed->chip, chipIt);
      chipIt.begin();
      if ( chipIt.valid() ) {
	if ( ntot < DIM_NFED ) {
	  fed_tot[ntot] = fed->id;
	  ntot++;
	} 
	else 
	  aframe_mess(routine,GRR_SWARNING,
		      "FED overflow, found more than %d", DIM_NFED ); 
      }
    } 
  } 

  /* check for faulty and missing FEDs and dump them */
  if ( event > 0 && ITR_silraw_iol.fed_statistics_period.value > 0 )
    if ( event%ITR_silraw_iol.fed_statistics_period.value == 0 ) {
      nok = nfaulty = nmissing = 0;

      /* loop over FEDs just to get the faulty ones */
      for ( ifed=0; ifed<nfed; ifed++ ) {
	imp_slbsub = etb_fed[ifed].hdr;
	if (   ITR_SILRAW_FED_RAW(imp_slbsub.recid)
	    || ITR_SILRAW_FED_SPARSE(imp_slbsub.recid) ) {
	  fed_ok[nok] = (imp_slbsub.fedid & 0xFF0F);
	  nok++;
	}
	else{ 
	  fed_faulty[nfaulty] = (imp_slbsub.fedid & 0xFF0F);
	  nfaulty++;
	}
      }

      /* Now check for missing ones */
      for ( i=0; i<ntot; i++ ) {
	found = 0;
	for ( j=0; j<nok; j++ )
	  if ( fed_tot[i] == fed_ok[j] )
	    found = 1;
	for ( j=0; j<nfaulty; j++ )
	  if ( fed_tot[i] == fed_faulty[j] )
	    found = 1;
	if ( !found ) {
	  fed_missing[nmissing] = fed_tot[i];
	  nmissing++;
	}
      }
	
      /* Finally print out the statistics */
      aframe_mess(routine,GRR_SINFO, "FED stats, event %d: %d nfed, %d total, %d ok, %d faulty, %d missing",
		  event, nfed, ntot, nok, nfaulty, nmissing );
      if ( nfaulty ) {
	sprintf(mess, "    Faulty FED(s) :");
	for ( i=0; i<nfaulty; i++ )
	  sprintf(mess, "%s %4x", mess, fed_faulty[i] );
	aframe_mess(routine,GRR_SINFO,mess);
      }
      if ( nmissing ) {
	sprintf(mess, "    Missing FED(s):");
	for ( i=0; i<nmissing; i++ )
	  /* printf(mess, "%s %4x", mess, fed_missing[i] );*/
	  printf("%6x",fed_missing[i]);
	/*aframe_mess(routine,GRR_SINFO,mess);*/
      }
      printf("\n");
      
    }
	


  /* count events */
  event++;


  return GRR_OK;
}

/* ----------------------------------------------------------------
 *
 * Fill ITR_Silraw_Event_t from FED blocks
 *
 * Inputs :  FED_HDR *fed : Pointer to first FED header
 *           int     nfed : number of FEDs
 *
 * ---------------------------------------------------------------- */
int ITR_silraw_get_fed( FED_HDR *fed, int nfed )
{
  int        ifed, rc;
  FED_HDR    imp_fed;    /* FED header in imported format */
  int        cellshort, celllong;
  int        error;

  ITR_Silraw_Event_t *silraw;

  char *routine = "ITR_silraw_get_fed";

  /* ---------------------------------------------------------- */

  /* fetch ITR_Silraw_Event_t */
  silraw = ITR_silraw_fetch_event();
  if ( !silraw )
    return GRR_FISHY;

  if ( ITR_silraw_iol.debug_mode.value )
    aframe_mess(routine,GRR_SINFO, "Raw data in FED format.");

  /* set counts to 0 */
  silraw->nfed    = 0;
  silraw->nchip   = 0;
  silraw->nsparse = 0;

  /* loop over FEDs */
  error = 0;
  for ( ifed=0; ifed<nfed; ifed++ ) {
    ESP_get_fed_hdr( fed, &imp_fed );

    rc = ITR_silraw_fetch_fed_from_fed( fed ); 
    if ( rc != GRR_OK ) {
      aframe_mess(routine,GRR_SERROR, "Could not read %d-th FED.", ifed);
      error++;
    }
    fed = (FED_HDR*) ( (char*) fed + imp_fed.length );
  }
  if ( error )
    return GRR_FISHY;

  /* fill ITR_Silraw_header_t from last FED header */
  silraw->header.fltevt = (int) imp_fed.fltevt;
  silraw->header.sltevt = 0;
  silraw->header.bxid   = (int) imp_fed.bxid;
  rc = ITR_silraw_fetch_cell_from_data( &cellshort, &celllong );
  if (rc != GRR_OK )
    return rc;
  /*VB*/
  silraw->header.cellshort = (unsigned short) cellshort;
  silraw->header.celllong  = (unsigned short) celllong;
  silraw->header.flag = 0;    /*for now */

  /* No daq event blocks known */
  silraw->daq        = 0;
     
  /* dump debug output if requested */
  if ( ITR_silraw_iol.debug_mode.value )
    aframe_mess(routine,GRR_SINFO,
		"Cellshort = %d,long=%d, FLT-EVT = %d, BX ID = %d",
		silraw->header.cellshort,silraw->header.cellshort, 
		silraw->header.fltevt,silraw->header.bxid );

  return GRR_OK;
}

/* ----------------------------------------------------------------
 *
 * Fill ITR_Silraw_Event_t from SLB sub-header blocks
 *
 * Inputs :  SLB_SUBHDR *slbsub : Pointer to first SLB sub-header
 *           int        nfed    : number of FEDs
 *
 * ---------------------------------------------------------------- */
int ITR_silraw_get_slbsub( SLB_SUBHDR *slbsub, int nfed )
{
  int        ifed, rc;
  SLB_SUBHDR imp_slbsub;    /* SLB sub-header in imported format */
  int        cellshort, celllong;
  int        error;

  ITR_Silraw_Event_t *silraw;

  char *routine = "ITR_silraw_get_slbsub";

  /* ---------------------------------------------------------- */

  /* fetch ITR_Silraw_Event_t */
  silraw = ITR_silraw_fetch_event();
  if ( !silraw )
    return GRR_FISHY;

  if ( ITR_silraw_iol.debug_mode.value )
    aframe_mess(routine,GRR_SINFO, "Raw data in SLB sub header format.");

  /* set counts to 0 */
  silraw->nfed    = 0;
  silraw->nchip   = 0;
  silraw->nsparse = 0;

  /* loop over FEDs */
  error = 0;
  for ( ifed=0; ifed<nfed; ifed++ ) {
    ESP_get_slb_subhdr( slbsub, &imp_slbsub );

    rc = ITR_silraw_fetch_fed_from_slbsub( slbsub ); 
    if ( rc != GRR_OK ) {
      aframe_mess(routine,GRR_SERROR, "Could not read %d-th FED.", ifed);
      error++;
    }      
    slbsub = (SLB_SUBHDR*) ( (char*) slbsub + imp_slbsub.length );
  }
  if ( error )
    return GRR_FISHY;

  /* fill ITR_Silraw_header_t from last FED header */
  silraw->header.fltevt = 0;
  silraw->header.sltevt = 0;
  silraw->header.bxid   = (int) imp_slbsub.bxid;
  rc = ITR_silraw_fetch_cell_from_data( &cellshort, &celllong );
  if (rc != GRR_OK )
    return rc;
  /*VB*/
  silraw->header.cellshort = (unsigned short) cellshort;
  silraw->header.celllong  = (unsigned short) celllong;
  silraw->header.flag = 0;    /*for now */

  /* No daq event blocks known */
  silraw->daq        = 0;
     
  /* dump debug output if requested */
  if ( ITR_silraw_iol.debug_mode.value )
    aframe_mess(routine,GRR_SINFO,
		"Cellshort = %d,long=%d, FLT-EVT = %d, BX ID = %d",
		silraw->header.cellshort,silraw->header.celllong,
		silraw->header.fltevt,silraw->header.bxid );

  return GRR_OK;
}

/* ----------------------------------------------------------------
 *
 * Fills one single FED block from ETB_fed structure into ITR_Silraw_Event_t
 *
 * Inputs : ETB_fed *etb_fed : Pointer to ETB_fed structure
 *
 * ---------------------------------------------------------------- */
int ITR_silraw_fetch_fed_from_etbfed( ETB_fed *etb_fed )
{
  int        ifed, rc, sloffset;
  int        nchip;  /* number of read-out chips per FADC */
  short      superlayer;

  ITR_Silraw_Event_t *silraw;

  char  *routine = "ITR_silraw_fetch_fed_from_etbfed";
 
  /* ---------------------------------------------------------- */
    
  /* fetch ITR_Silraw_Event_t */
  silraw = ITR_silraw_fetch_event();
  if ( !silraw )
    return GRR_FISHY;

 /* Update total FED count */
  silraw->nfed++;
  ifed = silraw->nfed - 1;

  /* fill FED structure in ITR_Silraw_Event_t
     ------------------------------------ */
  silraw->fed[ifed].fedid      = etb_fed->hdr.fedid;
  silraw->fed[ifed].recid      = etb_fed->hdr.recid;
  silraw->fed[ifed].length     = etb_fed->hdr.length;
  silraw->fed[ifed].header     = (char*) &(etb_fed->hdr);
  silraw->fed[ifed].headersize = sizeof(SLB_SUBHDR);
  silraw->fed[ifed].data       = (char*) etb_fed->data;
  silraw->fed[ifed].datasize   = etb_fed->hdr.length - sizeof(SLB_SUBHDR);
  silraw->fed[ifed].tail       = NULL;
  silraw->fed[ifed].tailsize   = 0;

  superlayer = ITR_silraw_get_superlayer(silraw->fed[ifed].fedid);
  if (superlayer == 0) {printf("%s: cant find superlayer\n",routine);return GRR_FISHY;}
  silraw->fed[ifed].superlayer = superlayer;
  sloffset = ITR_silraw_get_datavalidoffset(superlayer);
  if (sloffset == -1) {printf("%s: cant find offset\n",routine);return GRR_FISHY;}
  

  nchip                      = NCHIPFEDHELIX;
  silraw->fed[ifed].datatail = (char*) etb_fed->data
                             + ( nchip*(NDATACHAN+NCELLCHAN) + NOFFSETCHAN ) * 1;
  silraw->fed[ifed].offset   = ( ITR_silraw_iol.datavalid_jitter_offset.value ) % 4
                             + sloffset;
 

  if ( ITR_silraw_iol.debug_mode.value )
    aframe_mess(routine,GRR_SINFO,
		"FED  %d: id = 0x%x, length = %d, datasize = %d, "
		"offset = %d, datatail = 0x%x, nchip = %d",
		ifed, silraw->fed[ifed].fedid, silraw->fed[ifed].length,
		silraw->fed[ifed].datasize, silraw->fed[ifed].offset,
		(int) *(silraw->fed[ifed].datatail), nchip );

  /* Now fill chip structures in ITR_Silraw_Event_t
     ------------------------------------------ */
  rc = ITR_silraw_fetch_chips_from_fed( ifed, nchip );
  if ( rc != GRR_OK )
    return rc;

  return GRR_OK;
}

/* ----------------------------------------------------------------
 *
 * Fills one single FED block from SLB subheader into ITR_Silraw_Event_t
 *
 * Inputs : SLB_SUBHDR *slb_subhdr : Pointer to SLB sub header
 *
 * ---------------------------------------------------------------- */
int ITR_silraw_fetch_fed_from_slbsub( SLB_SUBHDR *slb_subhdr )
{
  ETB_fed  etb_fed;

  /* char  *routine = "ITR_silraw_fetch_fed_from_slbsub"; */

  /* ---------------------------------------------------------- */

  /* import SLB sub header to ETB_fed structure and fetch fed from there */
  ESP_get_slb_subhdr( slb_subhdr, &etb_fed.hdr );
  etb_fed.slb  = NULL;
  etb_fed.data = (void*) (slb_subhdr + 1);

  return ITR_silraw_fetch_fed_from_etbfed( &etb_fed );
}

/* ----------------------------------------------------------------
 *
 * Fills one single FED block from FED header into ITR_Silraw_Event_t
 *
 * Inputs : FED_HDR *fed_hdr : Pointer to FED header
 *
 * ---------------------------------------------------------------- */
int ITR_silraw_fetch_fed_from_fed( FED_HDR *fed_hdr )
{
  short      fedid, recid, superlayer;
  short      length;
  int        ifed, rc, sloffset;
  int        nchip;  /* number of read-out chips per FADC */

  FED_HDR    imp_fed_hdr;  /* imported FED header */

  ITR_Silraw_Event_t *silraw;

  char  *routine = "ITR_silraw_fetch_fed_from_fed";

  /* ---------------------------------------------------------- */

  /* fetch ITR_Silraw_Event_t */
  silraw = ITR_silraw_fetch_event();
  if ( !silraw )
    return GRR_FISHY;

  /* get SLB sub header */
  ESP_get_fed_hdr( fed_hdr, &imp_fed_hdr );
  length  = imp_fed_hdr.length;
  fedid   = imp_fed_hdr.fedid;
  recid   = imp_fed_hdr.recid;

  /* Update total FED count */
  silraw->nfed++;
  ifed = silraw->nfed - 1;

  /* fill FED structure in ITR_Silraw_Event_t
     ------------------------------------ */
  silraw->fed[ifed].fedid      = fedid;
  silraw->fed[ifed].recid      = fedid;
  silraw->fed[ifed].length     = length;
  silraw->fed[ifed].header     = (char*) fed_hdr;
  silraw->fed[ifed].headersize = sizeof(FED_HDR);
  silraw->fed[ifed].data       = (char*) fed_hdr + sizeof(FED_HDR);
  silraw->fed[ifed].datasize   = length - sizeof(FED_HDR) - sizeof(FED_TAIL);
  silraw->fed[ifed].tail       = (char*) fed_hdr + length - sizeof(FED_TAIL);
  silraw->fed[ifed].tailsize   = sizeof(FED_TAIL);

  superlayer = ITR_silraw_get_superlayer(fedid);
  if (superlayer == 0) return GRR_FISHY;
  silraw->fed[ifed].superlayer = superlayer;
  sloffset = ITR_silraw_get_datavalidoffset(superlayer);
  if (sloffset == -1) return GRR_FISHY;


  nchip                      = NCHIPFEDHELIX;
  silraw->fed[ifed].datatail = (char*) fed_hdr + sizeof(FED_HDR)
                             + ( nchip*(NDATACHAN+NCELLCHAN) + NOFFSETCHAN ) + 1;
  silraw->fed[ifed].offset   = ( ITR_silraw_iol.datavalid_jitter_offset.value ) % 4
                             + sloffset;

  if ( ITR_silraw_iol.debug_mode.value )
    aframe_mess(routine,GRR_SINFO,
		"FED  %d: id = 0x%x, length = %d, datasize = %d, "
		"offset = %d,  datatail = 0x%x, nchip = %d",
		ifed, silraw->fed[ifed].fedid, silraw->fed[ifed].length,
		silraw->fed[ifed].datasize, silraw->fed[ifed].offset,
		(int) *(silraw->fed[ifed].datatail), nchip );

  /* Now fill chip structures in ITR_Silraw_Event_t
     ------------------------------------------ */

  rc = ITR_silraw_fetch_chips_from_fed( ifed, nchip );
  if ( rc != GRR_OK )
    return rc;

  return GRR_OK;
}


/* ----------------------------------------------------------------
 *
 * Fills all chips from one FED into ITR_Silraw_Event_t
 *
 * Inputs : int ifed  : FED # in ITR_Silraw_Event_t (NOT FED ID!)
 *          int nchip : Number of chips of FED 'ifed'
 *
 * ---------------------------------------------------------------- */
int ITR_silraw_fetch_chips_from_fed( int ifed, int nchip )
{
  int   i, ichip;
  int   offset_bits;

  ITR_Silraw_Event_t *silraw;

  char  *routine = "ITR_silraw_fetch_chips_from_fed";

  /* ---------------------------------------------------------- */

  /* fetch ITR_Silraw_Event_t */
  silraw = ITR_silraw_fetch_event();
  if ( !silraw )
    return GRR_FISHY;

  /* Now fill chip structures in ITR_Silraw_Event_t */
  for ( i=0; i<nchip; i++) { 

    /* get current chip # */
    ichip = silraw->nchip;

    silraw->chip[ichip].fedid = silraw->fed[ifed].fedid;
    silraw->chip[ichip].recid = silraw->fed[ifed].recid;
    silraw->chip[ichip].flag  = 0;  /* for now */

    /* Helix - be careful: The offset arithmetic
                           only works with 8 and 12 bit data! */
 
    silraw->chip[ichip].chipid   = ( silraw->chip[ichip].fedid & 0xFF0F )
	                              + ( i << 4 );
    silraw->chip[ichip].length   =  NDATACHAN + NCELLCHAN ;
    offset_bits                  = silraw->fed[ifed].offset * 8;
    silraw->chip[ichip].offset   = ( offset_bits%8 == 0 ) ? 0 : 1 ;
       
    silraw->chip[ichip].data     =  silraw->fed[ifed].data
 	                             + i * silraw->chip[ichip].length
	                             + silraw->fed[ifed].offset
                                     - silraw->chip[ichip].offset 
                                     - silraw->chip[ichip].offset;
    silraw->chip[ichip].celldata = silraw->chip[ichip].data
                                 + NDATACHAN;
 

    if ( ITR_silraw_iol.debug_mode.value )
      aframe_mess(routine,GRR_SINFO,
		  "Chip %d: id = 0x%x, flag = %d, offset wrt. FED = %d",
		  ichip, silraw->chip[ichip].chipid, silraw->chip[ichip].flag,
		  silraw->chip[ichip].data - silraw->fed[ifed].data );

    /* update total chip count if not included in bad-chip-list */
    if ( !ITR_silraw_bad_chip( silraw->chip[ichip].chipid ) )
      silraw->nchip++;
    else if ( ITR_silraw_iol.debug_mode.value )
      aframe_mess(routine,GRR_SINFO,
		  "  -----> Included in list of bad chips !");
      
  }

  return GRR_OK;
}

/* ================================================================

   ROUTINES TO ACCESS THE EVENT STRUCTURE

   ================================================================ */

int ITR_silraw_get_major( int *major )
{
  /* fetch ITR_Silraw_Event_t */
  ITR_Silraw_Event_t *silraw;
  silraw = ITR_silraw_fetch_event();
  if ( !silraw )
    return GRR_FISHY;

  *major = silraw->header.major;

  return GRR_OK;
}

int ITR_silraw_get_minor( int *minor )
{
  /* fetch ITR_Silraw_Event_t */
  ITR_Silraw_Event_t *silraw;
  silraw = ITR_silraw_fetch_event();
  if ( !silraw )
    return GRR_FISHY;

  *minor = silraw->header.minor;

  return GRR_OK;
}

int ITR_silraw_get_evtime( int *evtime )
{
  /* fetch ITR_Silraw_Event_t */
  ITR_Silraw_Event_t *silraw;
  silraw = ITR_silraw_fetch_event();
  if ( !silraw )
    return GRR_FISHY;

  *evtime = silraw->header.evtime;

  return GRR_OK;
}


int ITR_silraw_get_cellshort( int *cellshort)
{
  /*fetch ITR_Silraw_Event_t*/
  ITR_Silraw_Event_t *silraw;
  silraw = ITR_silraw_fetch_event();
  if (!silraw)
    return GRR_FISHY;

  *cellshort = silraw->header.cellshort;

  return GRR_OK;
}

int ITR_silraw_get_celllong(int *celllong)
{
  /*fetch ITR_Silraw_Event_t*/
  ITR_Silraw_Event_t *silraw;
  silraw = ITR_silraw_fetch_event();
  if(!silraw)
    return GRR_FISHY;

  *celllong = silraw->header.celllong;

  return GRR_OK;
}

int ITR_silraw_get_fltevt( int *fltevt )
{
  /* fetch ITR_Silraw_Event_t */
  ITR_Silraw_Event_t *silraw;
  silraw = ITR_silraw_fetch_event();
  if ( !silraw )
    return GRR_FISHY;

  *fltevt = silraw->header.fltevt;

  return GRR_OK;
}

int ITR_silraw_get_sltevt( int *sltevt )
{
  /* fetch ITR_Silraw_Event_t */
  ITR_Silraw_Event_t *silraw;
  silraw = ITR_silraw_fetch_event();
  if ( !silraw )
    return GRR_FISHY;

  *sltevt = silraw->header.sltevt;

  return GRR_OK;
}

int ITR_silraw_get_bxid( int *bxid )
{
  /* fetch ITR_Silraw_Event_t */
  ITR_Silraw_Event_t *silraw;
  silraw = ITR_silraw_fetch_event();
  if ( !silraw )
    return GRR_FISHY;

  *bxid = silraw->header.bxid;

  return GRR_OK;
}

int ITR_silraw_get_nchip( int *nchip )
{
  /* fetch ITR_Silraw_Event_t */
  ITR_Silraw_Event_t *silraw;
  silraw = ITR_silraw_fetch_event();
  if ( !silraw )
    return GRR_FISHY;

  *nchip = silraw->nchip;

  return GRR_OK;
}

int ITR_silraw_get_data_chip( int chip, char **data )
{
  char *routine = "ITR_silraw_get_data_chip";

  ITR_Silraw_Event_t *silraw;

  *data = NULL;

  silraw = ITR_silraw_fetch_event();
  if ( !silraw )
    return GRR_FISHY;

  if ( chip >= silraw->nchip || chip < 0 ) {
    aframe_mess(routine,GRR_SERROR,
		"Invalid chip # %d (should be >0 and <%d)!",
		chip, silraw->nchip );
    return GRR_FISHY;
  }    

  *data = silraw->chip[chip].data;

  return GRR_OK;
}

int ITR_silraw_get_data_chipid( int chipid, char **data )
{
  int ichip;

  ITR_Silraw_Event_t *silraw;

  *data = NULL;

  silraw = ITR_silraw_fetch_event();
  if ( !silraw )
    return GRR_FISHY;

  for ( ichip=0; ichip<silraw->nchip; ichip ++ )
    if ( chipid == silraw->chip[ichip].chipid )
      *data = silraw->chip[ichip].data;

  return GRR_OK;
}

int ITR_silraw_get_nfed( int *nfed )
{
   /* fetch ITR_Silraw_Event_t */
  ITR_Silraw_Event_t *silraw;
  silraw = ITR_silraw_fetch_event();
  if ( !silraw )
    return GRR_FISHY;

  *nfed = silraw->nfed;

  return GRR_OK;
}

int ITR_silraw_get_nchan( int *nchan )
{
  *nchan = NDATACHAN;

  return GRR_OK;
}

int ITR_silraw_get_nsparse( int *nsparse )
{
  /* fetch ITR_Silraw_Event_t */
  ITR_Silraw_Event_t *silraw;
  silraw = ITR_silraw_fetch_event();
  if ( !silraw || (nsparse == NULL) )
    return GRR_FISHY;

  *nsparse = silraw->nsparse;

  return GRR_OK;
}

int ITR_silraw_get_chipid( int chip, int *chipid )
{
  char *routine = "ITR_silraw_get_chipid";

  ITR_Silraw_Event_t *silraw;

  *chipid = 0;

  silraw = ITR_silraw_fetch_event();
  if ( !silraw )
    return GRR_FISHY;

  if ( chip >= silraw->nchip || chip < 0 ) {
    aframe_mess(routine,GRR_SERROR,
		"Invalid chip # %d (should be >0 and <%d)!",
		chip, silraw->nchip );
    return GRR_FISHY;
  }    

  *chipid = silraw->chip[chip].chipid;

  return GRR_OK;
}

int ITR_silraw_get_pipelen(int chip, int *pipelen)
{
  char *routine = "ITR_silraw_get_pipelen";

  /* fetch ITR_Silraw_Event_t */
  ITR_Silraw_Event_t *silraw;
  silraw = ITR_silraw_fetch_event();
  if (!silraw )
    return GRR_FISHY;

  if (chip >= silraw->nchip || chip < 0){
    aframe_mess(routine,GRR_SERROR,
		"Invalid chip # %d (should be >0 and <%d)!",
		chip,silraw->nchip);
    return GRR_FISHY;
  }

  *pipelen = silraw->chip[chip].pipelen;
  return GRR_OK;
}

EVT_DATA *ITR_silraw_get_daqdata()
{
  /* fetch ITR_Silraw_Event_t */
  ITR_Silraw_Event_t *silraw;
  silraw = ITR_silraw_fetch_event();
  if ( !silraw )
    return 0;
  if ( !silraw->daq )
    return 0;

  return &(silraw->daq->data);
}


/* ================================================================

   OTHER UTILITY ROUTINES

   ================================================================ */

/* ----------------------------------------------------------------
 *
 * fetches the cell number from raw data (in ITR_Silraw_Event_t):
 *
 * HELIX:  cell number has to be determined from digitized analog data:
 *         = first 8 byte after channel digitizations 
 *         Remember that Helix cell bits are 'double-inverted', i.e.:
 *          - The first ADC count is the least significant bit
 *          - High ADC counts are 0's, low ADC counts are 1's
 *
 * ---------------------------------------------------------------- */
int ITR_silraw_fetch_cell_from_data( int *cellshort, int *celllong )
{
  int    cellsumshort[NCELLCHAN], cellsumlong[NCELLCHAN];
  int    bitmap[NCELLCHAN+1];
  float  max, min, average;
  int    i, j, ichip, rc, diff;
  int    sparse_cellshort, sparse_celllong, chipmin,chipmax;

  static int error_prescale = 0;

  ITR_Silraw_Event_t *silraw;

  char  mess[256];
  char  *routine = "ITR_silraw_fetch_cell_from_data";
  /*LedaTableIterator( ITR_Chip, chipIt );*/
  /* ---------------------------------------------------------- */

  /* Assume the worst case */
  *cellshort = -1;
  *celllong  = -1;

  /* fetch ITR_Silraw_Event_t */
  silraw = ITR_silraw_fetch_event();
  if ( !silraw )
    return GRR_FISHY;

  /* Do we have cell number from chips and/or sparse data ? */ 
  if ( silraw->nchip <= 0 && silraw->nsparse <= 0 && error_prescale % 100) {
    aframe_mess(routine,GRR_SERROR,
		"Nchip = 0 and no sparsified data ==> cell # set to 0! (%d times so far)",
		error_prescale + 1);
    *cellshort = 0;
    *celllong  = 0;
    
    error_prescale ++;
    return GRR_FISHY;
  }

  /* get cell number, if Nchip > 0 */
  if ( silraw->nchip > 0 ) {

    /* get cell # 
       ----------*/
  
    /* Loop over cell # data and average over them */
    for ( i=0; i<NCELLCHAN; i++){
      cellsumshort[i] = 0;
      cellsumlong[i]  = 0;
    }

    for ( ichip=0; ichip<silraw->nchip; ichip++) {
      /*if ( silraw->chip[ichip].offset > 1 || silraw->chip[ichip].offset < 0 ) {
	aframe_mess(routine,GRR_SERROR,"silraw->chip[%d].offset = %d not allowed!",
		    ichip, silraw->chip[ichip].offset );
	return GRR_FISHY;
      }*/ /*VB*/
      rc = ITR_silraw_reformat_data( silraw->chip[ichip].celldata,
				     NCELLCHAN+silraw->chip[ichip].offset,
				     bitmap );
      if ( rc != GRR_OK)
	return rc;

      /* check for saturated cell # */ 
      if (   bitmap[ silraw->chip[ichip].offset     ] >= 0xF0
	     && bitmap[ silraw->chip[ichip].offset + 1 ] >= 0xF0
	     && bitmap[ silraw->chip[ichip].offset + 2 ] >= 0xF0
	     && bitmap[ silraw->chip[ichip].offset + 3 ] >= 0xF0
	     && bitmap[ silraw->chip[ichip].offset + 4 ] >= 0xF0
	     && bitmap[ silraw->chip[ichip].offset + 5 ] >= 0xF0
	     && bitmap[ silraw->chip[ichip].offset + 6 ] >= 0xF0
	     && bitmap[ silraw->chip[ichip].offset + 7 ] >= 0xF0 ) {
	if ( ITR_silraw_iol.debug_mode.value ) {
	  sprintf(mess,"Cell# saturated for chip %4x:",
		  silraw->chip[ichip].chipid);
	  for ( j=0; j<8; j++ )
	    sprintf( mess, "%s %x", mess,
		     bitmap[ silraw->chip[ichip].offset + j ]);
	  aframe_mess(routine,GRR_SINFO,mess);
	}
	continue;
      }
      
      chipmin=255;
      chipmax=0;

      if (silraw->chip[ichip].pipelen == NCELLHELIXSHORT)
	for ( i=0; i<NCELLCHAN; i++) { /* LSB <-> MSB for cellnumber! */
	  cellsumshort[NCELLCHAN-i-1] += bitmap[ i + silraw->chip[ichip].offset ];
      	  if ( bitmap[ i + silraw->chip[ichip].offset ] < chipmin )   
	    chipmin = bitmap[ i + silraw->chip[ichip].offset ];
	  if ( bitmap[ i + silraw->chip[ichip].offset ] > chipmax )   
	    chipmax = bitmap[ i + silraw->chip[ichip].offset ];
	}

  
      if (silraw->chip[ichip].pipelen == NCELLHELIXLONG)
	for ( i=0; i<NCELLCHAN; i++){  /* LSB <-> MSB for cellnumber! */
	  cellsumlong[NCELLCHAN-i-1] += bitmap[ i + silraw->chip[ichip].offset ];
	  if ( bitmap[ i + silraw->chip[ichip].offset ] < chipmin )   
	    chipmin = bitmap[ i + silraw->chip[ichip].offset ];
	  if ( bitmap[ i + silraw->chip[ichip].offset ] > chipmax )   
	    chipmax = bitmap[ i + silraw->chip[ichip].offset ];
	}

      /*  printf (" CHIP %d\t%d\t%d\n", ichip, chipmin, chipmax);*/
      diff = chipmax-chipmin;
      silraw->chip[ichip].cellnumheight = diff;

   }/*end of loop over chips*/
    

    /*VB */
    if ( ITR_silraw_iol.debug_mode.value ) {
      printf("Found following sums of cell number data:\n ");
      for ( i=0; i<NCELLCHAN; i++)
	printf(" \t Bit %1d:", i);
      printf("\n short:");
      for ( i=0; i<NCELLCHAN; i++)
	printf(" %6x", cellsumshort[i]);
      printf("\n long: ");
      for ( i=0; i<NCELLCHAN; i++)
	printf(" %6x", cellsumlong[i]);
      printf("\n");
    }      

    /* find max,min value and center point */
    min = 256 * silraw->nchip ;
    max = 0.;
    for ( i=0; i<NCELLCHAN; i++ ) {
      if ( cellsumshort[i] > max )    max = cellsumshort[i];
      if ( cellsumshort[i] < min )    min = cellsumshort[i];
      if ( cellsumlong[i]  > max )    max = cellsumlong[i];
      if ( cellsumlong[i]  < min )    min = cellsumlong[i];
    }
    average = ( max + min ) / 2.;
    
    /* now get cell #s */
    *cellshort = 0;
    *celllong  = 0;
    for ( i=0; i<NCELLCHAN; i++ ) {
      *cellshort = *cellshort << 1;
      if ( cellsumshort[i] < average )  /* ADC counts are 'inverted' bits! */
	(*cellshort)++;
      *celllong = *celllong << 1;
      if ( cellsumlong[i] < average )  /* ADC counts are 'inverted' bits! */
	(*celllong)++;
    }

    /* check for "good" ADC counts */
    if ( max - min < 2 * silraw->nchip * ITR_silraw_iol.minadcdiff.value ){
      *cellshort = 0;
      *celllong  = 0;
    }
    else
      for ( i=0; i<NCELLCHAN; i++ )
	if ( fabs( cellsumshort[i] + cellsumlong[i] - average ) < ITR_silraw_iol.minadcdiff.value * silraw->nchip )
	  if (ITR_silraw_iol.debug_mode.value)
	    aframe_mess(routine,GRR_SWARNING,
			"Bit %d: ADC diff = %f, should be >%d or <-%d",
			i, (cellsumshort[i] + cellsumlong[i] - average)/silraw->nchip,
			ITR_silraw_iol.minadcdiff.value, ITR_silraw_iol.minadcdiff.value );
			
  }/*end of silraw->nchip loop*/
  if ( ITR_silraw_iol.debug_mode.value )
    aframe_mess(routine,GRR_SINFO,
	 	"Cell # from chip data (short): %d (long): %d", *cellshort, *celllong );

  /* check consistency with possible sparsified datasets
     --------------------------------------------------- */
  for ( i=0; i<silraw->nsparse; i++ ) {

    /* get sparse cell number and swap if necessary */
    sparse_celllong = *( (int*) silraw->sparse[i].data + 3 );
    sparse_cellshort= *( (int*) silraw->sparse[i].data + 2 );
    if ( sparse_celllong & 0xFFFF0000 )
      sparse_celllong = ESP_SWAP_INT(sparse_celllong);
    if ( sparse_cellshort & 0xFFFF0000 )
      sparse_cellshort = ESP_SWAP_INT(sparse_cellshort);

    if ( ITR_silraw_iol.debug_mode.value )
      aframe_mess(routine,GRR_SINFO,"Cell # from %d. sparsified data set: %d(short) %d(long)",
		  i, sparse_cellshort, sparse_celllong );
    
    /* put sparsified cell # into general cell #, if not yet filled */
    if ( *celllong == -1 ) 
      *celllong = sparse_celllong;
    if ( *cellshort == -1 ) 
      *cellshort = sparse_cellshort;

    /* check consistency */
    if ( (sparse_celllong != *celllong) || (sparse_cellshort != *cellshort) )
      aframe_mess(routine,GRR_SERROR,
		  "Cell # %d from %d. sparsified data set inconsistent with earlier cell # %d !",
		  sparse_celllong, i, *celllong );
  }

  return GRR_OK;
}

/* ----------------------------------------------------------------
 *
 * Convert the digitization information from 8/12 bit words to int.
 *
 * HELIX:   8 bit data words (2 nibbles)
 *
 * Inputs : char *data   : Pointer to first raw digitization
 *          int  ndata   : Number of digitizations
 *
 * Output : int  *output : int array with integer converted digitizations
 *
 * ---------------------------------------------------------------- */
int ITR_silraw_reformat_data(unsigned char *data, int ndata, int *output )
{
  
  int  i;

  for ( i=0; i<ndata; i++)
    output[i] = (int) data[i];

  return GRR_OK;
}


/* ----------------------------------------------------------------
 *
 * Get array of FED headers (=SLB sub-headers) for a particular SLB.
 * Functionality, input, and outputs are similar to
 * the ETB routines (like ETB_get_feds(...) ).
 *
 * ---------------------------------------------------------------- */
int ITR_silraw_etb_get_feds_from_slb( const SLB_HDR *slb_hdr, int max,
				  ETB_fed *feds, int *nfed )
{
  int            slblen, fedlen;
  char           *p1 = (char *) slb_hdr;
  char           *p2;
  static SLB_HDR imp_slb_hdr;
  SLB_HDR        ext_slb_hdr;
  SLB_TAIL       *slb_tail;
  SLB_SUBHDR     imp_fed_hdr;

  /* initialize return variables */
  *nfed = 0;

  /* import SLB header */
  ext_slb_hdr = *slb_hdr;
  ESP_get_slb_hdr( &ext_slb_hdr, &imp_slb_hdr );

  /* get the SLB block */
  slblen = imp_slb_hdr.length;
    
  /* Check the SLB block */
  if ( slblen < sizeof(SLB_HDR) + sizeof(SLB_TAIL) ) 
    return ETB_corrupted_length;
  slb_tail = (SLB_TAIL *) ( p1 + slblen - sizeof(SLB_TAIL));
  if (    slb_tail->recid  != slb_hdr->recid 
       || slb_tail->fltevt != slb_hdr->fltevt )
    return ETB_corrupted_slb;

  /* loop over FED blocks */
  p2 = (char *)(slb_hdr + 1);
  while ( p2 + sizeof(SLB_TAIL) < p1 + slblen) {
	  
    /* get the FED block */
    ESP_get_slb_subhdr( (SLB_SUBHDR *) p2, &imp_fed_hdr );
    fedlen = (int) imp_fed_hdr.length;
    if ( fedlen < sizeof(SLB_SUBHDR) )
      return ETB_corrupted_length;
    *nfed += 1;
    if ( *nfed <= max ){  /* copy the information */
      feds->hdr  = imp_fed_hdr;
      feds->slb  = &imp_slb_hdr;
      feds->data =(void *) ( ((SLB_SUBHDR*) p2) + 1);
      feds++;
    }
    p2 += fedlen;
  }

  if ( *nfed > max )
    return ETB_out_of_memory;

  return ETB_ok;
}


/* ================================================================

   ROUTINES TO DUMP OUT INFORMATION

   ================================================================ */

/* dump one event if requested */
int ITR_silraw_dump_event()
{
  /* dump the header ? */
  if ( ITR_silraw_iol.dump_header.value )
    ITR_silraw_dump_header();

  /* dump the chips ? */
  if ( ITR_silraw_iol.dump_chips.value )
    ITR_silraw_dump_chips();

  /* dump the feds ? */
  if ( ITR_silraw_iol.dump_feds.value )
    ITR_silraw_dump_feds();

  return GRR_OK;
}

/*---------------------------------------------------------------------------------------*/
/* dump chips */
int ITR_silraw_dump_chips()
{
  int i, j;

  ITR_Silraw_Event_t *silraw;

  /* fetch ITR_Silraw_Event_t */
  silraw = ITR_silraw_fetch_event();
  if ( !silraw )
    return GRR_FISHY;

  /* now dump the body */
  for ( i=0; i<silraw->nchip; i++) {
    printf("Chip%4d: id=0x%x, FEDid=0x%x, length=%d, offset=%d, flag=%d, pipelen=%d\n",
	   i, silraw->chip[i].chipid, silraw->chip[i].fedid,
	   silraw->chip[i].length, silraw->chip[i].offset, silraw->chip[i].flag, 
	   silraw->chip[i].pipelen);
    for ( j=0; j<silraw->chip[i].length; j++ ) {
      printf(" %2x", (unsigned char) silraw->chip[i].data[j] );
      if ( (j+1)%32 == 0 ) 
	printf("\n");
    }
    printf("\n");
  }

  return GRR_OK;
}

/*---------------------------------------------------------------------------------------*/
/* dump feds */
int ITR_silraw_dump_feds()
{
  int i, j;

  ITR_Silraw_Event_t *silraw;

  /* fetch ITR_Silraw_Event_t */
  silraw = ITR_silraw_fetch_event();
  if ( !silraw )
    return GRR_FISHY;

  /* dump the FEDs */
  for ( i=0; i<silraw->nfed; i++) {
    printf("FED%4d: FEDid=0x%x, length=%d, offset=%d\n",
	   i, silraw->fed[i].fedid, silraw->fed[i].length,
	   silraw->fed[i].offset );
    printf("Header: ");
    for ( j=0; j<silraw->fed[i].headersize; j++ )
      printf(" %2x", (unsigned char) silraw->fed[i].header[j] );
    printf("\n");
    printf("Data  : ");
    for ( j=0; j<silraw->fed[i].datasize; j++ ) {
      printf(" %2x", (unsigned char) silraw->fed[i].data[j] );
      if ( (j+1)%32 == 0 ) 
	printf("\n        ");
    }
    printf("\n");
    if ( silraw->fed[i].tailsize > 0 ) {
      printf("Tail  : ");
      for ( j=0; j<silraw->fed[i].tailsize; j++ )
	printf(" %2x", (unsigned char) silraw->fed[i].tail[j] );
      printf("\n");
    }
  }

  return GRR_OK;
}

/*---------------------------------------------------------------------------------------*/
/* dump the header */
int ITR_silraw_dump_header()
{
  ITR_Silraw_Event_t *silraw;

  /* fetch ITR_Silraw_Event_t */
  silraw = ITR_silraw_fetch_event();
  if ( !silraw )
    return GRR_FISHY;

  /* dump header */
  printf("ITR_Silraw_Event_t Header: Run#(major) = %d, Event#(minor) = %d\n", 
	 silraw->header.major, silraw->header.minor);
  printf("                       Nfed = %d, Nchip = %d\n", 
	 silraw->nfed, silraw->nchip);
  printf("                       FLTno = %d, BXid = %d\n", 
	 silraw->header.fltevt, silraw->header.bxid);
  printf("                       short cell # = %d, long cell # = %d\n",
	 silraw->header.celllong, silraw->header.cellshort);

  return GRR_OK;
}

/* dump sparsified data */
int ITR_silraw_dump_sparse()
{
  int i, j;

   ITR_Silraw_Event_t *silraw;

  /* fetch ITR_Silraw_Event_t */
  silraw = ITR_silraw_fetch_event();
  if ( !silraw )
    return GRR_FISHY;

 /* dump sparsified data (if available) */
  if ( silraw->nsparse == 0 ) {
    printf("No sparsified data in silraw.\n");
    return GRR_OK;
  }
  for ( i=0; i<silraw->nsparse; i++) {
    printf("Sparsified data set %4d: Length=%d, Record id=0x%x\n",
	   i, silraw->sparse[i].length, silraw->sparse[i].recid );
    printf("Data  : ");
    for ( j=0; j<silraw->sparse[i].length; j++ ) {
      printf(" %2x", (unsigned char) silraw->sparse[i].data[j] );
      if ( (j+1)%32 == 0 ) 
	printf("\n        ");
    }
    printf("\n");
  }

  return GRR_OK;
}



/*---------------------------------------------------------------------------------------*/
/* get superlayer */
short ITR_silraw_get_superlayer(short fedid)
{
  short superlayer = 0;

  LedaTableRow(ITR_Fed) *ledafed = ITR_Fed->get->key(fedid & 0xFF0F);

  if (ledafed != NULL) {
    LedaRelationIterator(ledafed->chip, chipIt);
    chipIt.begin();
    if (chipIt.valid()){
      LedaTableRow(ITR_Chip)    *ledachip    = chipIt.get();
      LedaTableRow(ITR_Chamber) *ledachamber = LedaRelation(ledachip->chamber);
      LedaTableRow(ITR_nede)    *ledanede    = LedaRelation(ledachamber->tonede);
      superlayer = ledanede->superl_db;
    }
  }
  
  return superlayer;
}

/* get datavalid offset */
int ITR_silraw_get_datavalidoffset(short superlayer)
{
  int offset=-1;
  switch (superlayer){
  case 1: offset = ITR_silraw_iol.datavalid_offset_sl1.value;
    break;
  case 2: offset = ITR_silraw_iol.datavalid_offset_sl2.value;
    break;
  case 3: offset = ITR_silraw_iol.datavalid_offset_sl3.value;
    break;
  case 4: offset = ITR_silraw_iol.datavalid_offset_sl4.value;
    break;  
  case 5: offset = ITR_silraw_iol.datavalid_offset_sl5.value;
    break;
  case 6: offset = ITR_silraw_iol.datavalid_offset_sl6.value;
    break;
  case 7: offset = ITR_silraw_iol.datavalid_offset_sl7.value;
    break;
  case 8: offset = ITR_silraw_iol.datavalid_offset_sl8.value;
    break;  
  case 9: offset = ITR_silraw_iol.datavalid_offset_sl9.value;
    break;
  case 10: offset = ITR_silraw_iol.datavalid_offset_sl10.value;
    break;  

  }

  return offset;
}


/* get datavalid offset */
int ITR_silraw_get_datavalidoffset_from_chip(short chipid)
{
  int offset=-1;
  int superlayer;

  LedaTableRow(ITR_Chip) *ledaChip;
  LedaTableRow(ITR_Chamber) *ledaChamber;
  LedaTableRow(ITR_nede) *ledaNede;

  ledaChip = ITR_Chip->get->key(chipid);
  if (ledaChip == NULL)
    return -1;

  ledaChamber = LedaRelation(ledaChip->chamber);
  if (ledaChamber == NULL)
    return -1;
  
  ledaNede = LedaRelation(ledaChamber->tonede);
  if (ledaNede == NULL)
    return -1;

  superlayer = ledaNede->superl_db;


  switch (superlayer){
  case 1: offset = ITR_silraw_iol.datavalid_offset_sl1.value;
    break;
  case 2: offset = ITR_silraw_iol.datavalid_offset_sl2.value;
    break;
  case 3: offset = ITR_silraw_iol.datavalid_offset_sl3.value;
    break;
  case 4: offset = ITR_silraw_iol.datavalid_offset_sl4.value;
    break;  
  case 5: offset = ITR_silraw_iol.datavalid_offset_sl5.value;
    break;
  case 6: offset = ITR_silraw_iol.datavalid_offset_sl6.value;
    break;
  case 7: offset = ITR_silraw_iol.datavalid_offset_sl7.value;
    break;
  case 8: offset = ITR_silraw_iol.datavalid_offset_sl8.value;
    break;  
  case 9: offset = ITR_silraw_iol.datavalid_offset_sl9.value;
    break;
  case 10: offset = ITR_silraw_iol.datavalid_offset_sl10.value;
    break;  

  }

  return offset;
}




int ITR_silraw_check_event()
/*-----------------------------------------------------------------------------
 * check/repair the event integrity
 * ================================
 * For all chips the column number read out from the HELIX is checked,
 * using as input the offset provided in ITR_silraw_iol.datavalid_offset.
 * In a first round only chips with id=0 are examined.
 * In repair mode the validity of the assumed offset is checked by the 
 * program, and in case it fails the test it is overwritten by an 
 * alternative offset-hypothesis. In non-repair mode the given offset 
 * is accepted unconditionally. Using the offset, the column number 
 * of all id=0 chips is determined. The most frequent number is then
 * taken as the nominal one. In a second iteration over all chips 
 * the column number obtained by using the input offset is compared
 * with the nominal one. In case of discrepancies the repair mode 
 * searches for alternative offsets that reproduce the nominal column
 * number. If the nominal column number is not found, then the chip 
 * is flagged as "bad": chip->flag = 1, otherwise the data pointer 
 * shifted by the proper offset. Finally also the nominal cell number
 * is put into the ITR_silraw header structure. 
 * ... and this all for helixes with the long and the short pipeline
 *
 * If the pipelinelength is false the chip is flagged as "error with 
 * pipelen": chip->flag = 2. Often this is the case when the chamber 
 * not exist but is in the database!   
 *
 * In case any of the ITR_silraw debug-flags is on some information 
 * concerning cell numbers and offsets is printed before return.
 *
 * ITR_silraw_check_event() should be called before calling marvin.
 *
 * Author: Michael Schmelling / 3-Sep-1998
 *---------------------------------------------------------------------------*/
{
  int           i, j, failed, nfail;
  int           nhitshort, nhitlong, nmissshort, nmisslong, n2flag=0;
  int           rc, nchip, chipid, offset;
  int           icellshort, icelllong, cellnumshort, cellnumlong,  pipelen, flag;
  unsigned char *chipdata;
  static   int  cellshort[NCELLHELIXSHORT], celllong[NCELLHELIXLONG], offs[NOFFS];
  static   int  ccell[DIM_NCHIP], coffs[DIM_NCHIP];

  ITR_Silraw_Event_t *silraw;
  char *routine = "ITR_silraw_check_event";

  /*---------------------------------------------------------------------------*/

  /* get the number of chips that were read out */

  if ((rc = ITR_silraw_get_nchip (&nchip)) != GRR_OK) return rc;
  if (nchip == 0) 
    return GRR_FISHY;
  
  /* get ITR_silraw event */
  silraw = ITR_silraw_fetch_event();
  if ( !silraw )
    return GRR_FISHY;

  /* initialize the cell count */ /*VB*/
  for(i=0; i<NCELLHELIXSHORT; i++) cellshort[i] = 0;
  for(i=0; i<NCELLHELIXLONG;  i++) celllong[i]  = 0;


  /*
   *  deal with the non-repair mode:
   *
   *   take offset as given in iol
   *   determine all cell numbers
   *   flag all chips that don't conform to the majority
   *   fill cell number in ITR_silraw-header
   *   maybe do some debug prints and return
   */

  if( !ITR_silraw_iol.repair_mode.value ) {

   
    /* find and count columnnumbers */
    for(i=0; i<nchip; i++) {
      chipdata = silraw->chip[i].data;
      pipelen  = silraw->chip[i].pipelen;
      flag     = silraw->chip[i].flag;
      /*VB*/
       if (pipelen == NCELLHELIXSHORT){
	icellshort = ITR_silraw_fit_cell(chipdata + 128);
	ccell[i]   = icellshort;
	cellshort[icellshort]++;
      }
      else if (pipelen == NCELLHELIXLONG){
	icelllong = ITR_silraw_fit_cell(chipdata + 128); 
	ccell[i]  = icelllong;
	celllong[icelllong]++;
      }
      else {
	if(ITR_silraw_iol.debug_mode.value)
	  aframe_mess(routine, GRR_SERROR, "chip: 0x%x, false pipelinelength %d",
		      silraw->chip[i].chipid, pipelen);
	silraw->chip[i].flag = 2; /*VB: this means false pipeline, 
				    often in this case the chamber does not exist*/
	n2flag++;
      }
    } /*end for(...nchip...)-loop*/
    

    /* find most frequent columnnumber */
    cellnumshort = cellnumlong = 0;
    /*VB:*/
    for(i=0;i<NCELLHELIXSHORT; i++) {
      if(cellshort[i]>cellshort[cellnumshort]) cellnumshort = i;
    }
    for(i=0;i<NCELLHELIXLONG;i++){
      if(celllong[i]>celllong[cellnumlong]) cellnumlong = i;
    }
    nhitshort = nhitlong = nmissshort = nmisslong = 0;
    for(i=0; i<nchip; i++) {
      pipelen  = silraw->chip[i].pipelen;
      flag     = silraw->chip[i].flag;
      if (pipelen == NCELLHELIXSHORT){
	if(ccell[i] == cellnumshort) {
	  nhitshort++;
	} else {
	  nmissshort++;
	  silraw->chip[i].flag = 1;
	}
      }
      else if (pipelen == NCELLHELIXLONG){
	if(ccell[i] == cellnumlong) {
	  nhitlong++;
	} else {
	  nmisslong++;
	  silraw->chip[i].flag = 1;
	}
      }
      else {
	if (flag != 2){
	  if(ITR_silraw_iol.debug_mode.value)
	    aframe_mess(routine, GRR_SINFO, "chip 0x%x, false pipelinelength %d",
			silraw->chip[i].chipid,pipelen);
	  silraw->chip[i].flag = 2; /*VB: this means false pipeline, 
		      often in this case the chamber does not exist*/
	  n2flag++;
	}
      }
    } 

    silraw->header.cellshort = cellnumshort;
    silraw->header.celllong  = cellnumlong;
    
    if( ITR_silraw_iol.dump_header.value ||
        ITR_silraw_iol.dump_chips.value  ||
        ITR_silraw_iol.dump_feds.value   ||
        ITR_silraw_iol.debug_mode.value    ) { 
      printf("*** event check performed in non-repair mode ***\n");
      printf("cell number for short pipeline is           %4d \n",cellnumshort);
      /* printf("nominal offset from iol  %4d \n",
              ITR_silraw_iol.datavalid_offset.value);*/
      printf("CORRECT offset used for  %4d chips\n",nhitshort);   
      printf("WRONG   offset used for  %4d chips\n",nmissshort);
      if(nmissshort>nhitshort)
      printf("---> using repair mode recommended <--- \n\n");  
      printf("cell number for long pipeline is           %4d \n",cellnumlong);
      /* printf("nominal offset from iol  %4d \n",
              ITR_silraw_iol.datavalid_offset.value);*/
      printf("CORRECT offset used for  %4d chips\n",nhitlong);   
      printf("WRONG   offset used for  %4d chips\n",nmisslong);
      printf("number of chips with false pipeline: %4d \n",n2flag);
      printf("          (exist these chambers ??)\n");
      if(nmisslong>nhitlong)
      printf("---> using repair mode recommended <--- \n");  
    }
  
    return GRR_OK;
  }


  /*
   *  deal with the repair mode: 
   *   
   *   first use all chips with id=0
   *   verify the offset
   *   determine cell numbers
   *   accept majority as the nominal cell number
   *   then loop over all chips and if needed try to adjust the offset
   *   flag chips that cannot be made to conform to the majority
   *   fill cell number in ITR_silraw-header
   *   maybe do some debug prints and return
   */

  /* 
   *  loop over all chips and get nominal cell number from chips with id=0 
   */

  /* search right offset at the beginning of data */
  for(i=0; i<nchip; i++) {
    offset   = ITR_silraw_get_datavalidoffset_from_chip (silraw->chip[i].chipid);
    chipdata = silraw->chip[i].data - offset;
    chipid   = silraw->chip[i].chipid & 0x00F0;
    pipelen  = silraw->chip[i].pipelen;
    flag     = silraw->chip[i].flag;
    if(chipid==0 && offset!=0) {
      if(ITR_silraw_check_offset(chipdata, offset)) {
	for(j=1; j<NOFFS; j++) {
          if(!ITR_silraw_check_offset(chipdata, j)) {
            offset = j; 
            break;
          } 
        }
      }
    }
    /* find and count columnnumbers */
    if(chipid == 0) {
      if (pipelen == NCELLHELIXSHORT){
	icellshort  = ITR_silraw_fit_cell(chipdata + offset + 128);
	ccell[i] = icellshort;
	coffs[i] = offset;
	cellshort[icellshort]++;
      }
      else if (pipelen == NCELLHELIXLONG){
	icelllong  = ITR_silraw_fit_cell(chipdata + offset + 128);
	ccell[i] = icelllong;
	coffs[i] = offset;
	celllong[icelllong]++;
      }
      else {
	if(ITR_silraw_iol.debug_mode.value)
	  aframe_mess (routine, GRR_SERROR, "chip 0x%x, false pipelinelength %d", 
		       chipid,pipelen);
	silraw->chip[i].flag = 2; /*VB: this means false pipeline, 
		    often in this case the chamber does not exist*/
	n2flag++;
      }
    }
  }  /*end for(... nchip...)-loop*/

  /* find most frequent columnnumber */
  cellnumshort = cellnumlong =  0;
  for(i=0; i<NCELLHELIXSHORT; i++) {
    if(cellshort[i]>cellshort[cellnumshort]) cellnumshort = i;
  }
  for(i=0; i<NCELLHELIXLONG; i++){
    if(celllong[i]>celllong[cellnumlong])    cellnumlong = i;
  }

  /*
   *  loop over all chips and try to find the correct cell number
   */

  
  for(i=0; i<NOFFS; i++) offs[i]=0;  /* zero offset counters   */
  nfail = nhitshort = nhitlong = nmissshort = nmisslong = 0;  /* zero success/failure statistics */

  for(i=0; i<nchip; i++) {
    failed = 1;
    chipdata = silraw->chip[i].data 
      - ITR_silraw_get_datavalidoffset_from_chip (silraw->chip[i].chipid);
    chipid   = silraw->chip[i].chipid & 0x00F0;
    pipelen  = silraw->chip[i].pipelen;
    if(pipelen == NCELLHELIXSHORT){ 
      if(chipid==0 && ccell[i]==cellnumshort) {
	offset = coffs[i];
	failed = 0;
	goto DONE;
      } else {
	offset = ITR_silraw_get_datavalidoffset_from_chip (silraw->chip[i].chipid);
	icellshort   = ITR_silraw_fit_cell(chipdata + offset + 128);
	if(icellshort == cellnumshort) {
	  failed = 0;
	  goto DONE; 
	} else {
	  for(j=1; j<NOFFS; j++) {
	    icellshort = ITR_silraw_fit_cell(chipdata + j + 128);
	    if(icellshort == cellnumshort) {
	      offset = j;
	      failed = 0;
	      goto DONE; 
	    }
	  } 
	}       
      }
    }
    else if (pipelen == NCELLHELIXLONG){
      if(chipid==0 && ccell[i]==cellnumlong) {
	offset = coffs[i];
	failed = 0;
	goto DONE;
      } else {
	offset = ITR_silraw_get_datavalidoffset_from_chip (silraw->chip[i].chipid);
	icelllong   = ITR_silraw_fit_cell(chipdata + offset + 128);
	if(icelllong == cellnumlong) {
	  failed = 0;
	  goto DONE; 
	} else {
	  for(j=1; j<NOFFS; j++) {
	    icelllong = ITR_silraw_fit_cell(chipdata + j + 128);
	    if(icelllong == cellnumlong) {
	      offset = j;
	      failed = 0;
	      goto DONE; 
	    }
	  } 
	}       
      } 
    }
    else { 
      if (flag != 2)
	{
	  if(ITR_silraw_iol.debug_mode.value)
	    aframe_mess(routine, GRR_SERROR, "chip: 0x%x, false pipelinelength %d\n", 
			chipid, pipelen);
	  silraw->chip[i].flag = 2; /*VB: this means false pipeline, 
		      often in this case the chamber does not exist*/
	  n2flag++;
	}
      failed = 0;
      goto DONE;
    }
    
    DONE:
    if(failed) {
      if (silraw->chip[i].flag != 2)
	silraw->chip[i].flag = 1;
      nfail++;
    } else {
      offs[offset]++;  /* count offsets */
      silraw->chip[i].data += 
	offset - ITR_silraw_get_datavalidoffset_from_chip (silraw->chip[i].chipid);
      /*new pointerposition*/
      if(offset == ITR_silraw_get_datavalidoffset_from_chip (silraw->chip[i].chipid)) {
	if (pipelen == NCELLHELIXSHORT) nhitshort++;
	else nhitlong++;
      } else {
	if(pipelen == NCELLHELIXSHORT) nmissshort++;
	else nmisslong++;
      }
    }
  } /*end for(...nchip...)-loop*/

  /* fill the pipeline column number */

  silraw->header.cellshort = cellnumshort;
  silraw->header.celllong  = cellnumlong;

  /* print diagnostics if any debugging is switched on */

  if( ITR_silraw_iol.dump_header.value ||
      ITR_silraw_iol.dump_chips.value  ||
      ITR_silraw_iol.dump_feds.value   ||
      ITR_silraw_iol.debug_mode.value  ||
      (ITR_silraw_iol.cell_check_output.value == 2 )) { 

      /* determine also the preferred offset */  
    
      offset = 0;
      for(i=0; i<NOFFS; i++) {
         if(offs[i]>offs[offset]) offset = i;
      }
      
      printf("*** event check performed in REPAIR mode ***\n");     
      printf("current cell number for short pipeline is             : %4d \n",cellnumshort); 
      printf("preferred offset is                : %4d \n",offset);
      /*  printf("nominal offset taken from iol is   : %4d \n",
              ITR_silraw_iol.datavalid_offset.value);*/
      printf("number of chips with nominal offset: %4d \n",nhitshort);
      printf("                 alternative offset: %4d \n",nmissshort);
      printf("current cell number for long pipeline is             : %4d \n",cellnumlong); 
      printf("number of chips with nominal offset: %4d \n",nhitlong);
      printf("                 alternative offset: %4d \n",nmisslong);
      printf("        offset determination failed: %4d \n",nfail);
      printf("number of chips with false pipeline: %4d \n",n2flag);
      printf("             (exist these chambers??)\n");
  }

   return GRR_OK;
}


/*-----------------------------------------------------------------------------
 * check cell number consistency for all chips
 * ===========================================
 * 
 * Author: Rainer Wanke / 1-Oct-1998
 * long and short pipeline: VB July 1999
 *---------------------------------------------------------------------------*/
int ITR_silraw_check_cell_consistency() 
{
  int    ifed, icell, i, nchip;
  int    linkfail;
  int    cell1[DIM_NFED], cell2[DIM_NFED], rc;
  /*VB  appearance count of each cell in event*/
  int    cellcountshort[NCELLHELIXSHORT],cellcountlong[NCELLHELIXLONG];
  int    bestcellshort,bestcelllong; 
  int    ncellmaxshort,ncellmaxlong; 

  static int shit_happens       = 0;
  static int shit_happens_count = 0;
  static int shit_happens_event = 0;

  ITR_Silraw_Event_t *silraw;

  time_t mytime;

  char *routine = "ITR_silraw_check_cell_consistency";

  /* ---------------------------------------------------------- */

  /* fetch ITR_Silraw_Event_t */
  silraw = ITR_silraw_fetch_event();
  if ( !silraw )
    return GRR_FISHY;

  /*VB*/
  if ((rc = ITR_silraw_get_nchip (&nchip)) != GRR_OK) return rc;
  if (nchip == 0) return GRR_FISHY;

  /* initialization */
  linkfail = 0;
  for ( icell = 0; icell<NCELLHELIXMAX; icell++ ){
    cellcountshort[icell] = 0; cellcountlong[icell] = 0;
  }

  /* print event number */
  if ( ITR_silraw_evno < 15 && ITR_silraw_iol.cell_check_output.value ) {
    aframe_mess(routine,GRR_SINFO,"----------------------------------");
    aframe_mess(routine,GRR_SINFO,"           EVENT %d", ITR_silraw_evno );
    aframe_mess(routine,GRR_SINFO,"----------------------------------");
  }

 
  /* loop over all feds (=links) */  
  for ( ifed=0; ifed<silraw->nfed; ifed++) {

    /*VB*/
    /*fill fed pipelen*/
  
    for(i=0;i<nchip;i++){
      if ((silraw->chip[i].fedid & 0xff0f) == (silraw->fed[ifed].fedid & 0xff0f) ){ 
	silraw->fed[ifed].pipelen = silraw->chip[i].pipelen;
	/*if(ITR_silraw_iol.debug_mode.value)
	  aframe_mess(routine,GRR_SINFO,"Fed 0x%x Pipeline %d",
	  silraw->fed[ifed].fedid,silraw->fed[ifed].pipelen);
	*/
      }
    }


    /* fill fedid and cell number arrays */    
    if ( ITR_silraw_evno == 1 ) {
      ITR_silraw_chipid[2*ifed]   = ( silraw->fed[ifed].fedid & 0xff0f );
      ITR_silraw_chipid[2*ifed+1] = ( silraw->fed[ifed].fedid & 0xff0f ) | 0x10;
    }
    cell1[ifed] = ITR_silraw_fit_cell( silraw->fed[ifed].data
				   + silraw->fed[ifed].offset + 128 );
    cell2[ifed] = ITR_silraw_fit_cell( silraw->fed[ifed].data
				   + silraw->fed[ifed].offset + 128 + 136 );
    if ( ITR_silraw_evno < 15 && ITR_silraw_iol.cell_check_output.value == 2 )
      aframe_mess(routine,GRR_SINFO, "link=%4d, fed=0x%x: cells are %3d %3d",
		  ifed, silraw->fed[ifed].fedid, cell1[ifed], cell2[ifed]);

    /* check and count cell numbers */
    /*VB: */
    if ( silraw->fed[ifed].pipelen == NCELLHELIXSHORT ){
      if ( cell1[ifed] < NCELLHELIXSHORT )
	cellcountshort[cell1[ifed]]++;
      else
	aframe_mess(routine,GRR_SINFO, "Short cell # %d out of range, link=%d, fed=0x%x, 1.helix",
		    cell1[ifed], ifed, silraw->fed[ifed].fedid ); 
      
      if ( cell2[ifed] < NCELLHELIXSHORT )
	cellcountshort[cell2[ifed]]++;
      else
	aframe_mess(routine,GRR_SINFO,"Short cell # %d out of range, link=%d, fed=0x%x, 2.helix",
		    cell2[ifed], ifed, silraw->fed[ifed].fedid ); 
    }
    
    else if ( silraw->fed[ifed].pipelen == NCELLHELIXLONG ){
      if ( cell1[ifed] < NCELLHELIXLONG )
	cellcountlong[cell1[ifed]]++;
      else
	aframe_mess(routine,GRR_SINFO, "Long cell # %d out of range, link=%d, fed=0x%x, 1.helix",
		    cell1[ifed], ifed, silraw->fed[ifed].fedid ); 
      
      if ( cell2[ifed] < NCELLHELIXLONG )
	cellcountlong[cell2[ifed]]++;
      else
	aframe_mess(routine,GRR_SINFO,"Long cell # %d out of range, link=%d, fed=0x%x, 2.helix",
		    cell2[ifed], ifed, silraw->fed[ifed].fedid ); 
    }
    
    else 
      {if(ITR_silraw_iol.debug_mode.value)
      aframe_mess (routine, GRR_SWARNING, "pipelinelength %d from FED 0x%x not valid!", 
		      silraw->fed[ifed].pipelen, silraw->fed[ifed].fedid);
    /*VB: FED markieren??Kammer vorhanden??*/
      }
  }/*end of fed-loop*/


  /*VB: aufgeteilt...*/
  /* Now get most preferred cell # */
  ncellmaxshort = 0;  ncellmaxlong = 0;
  bestcellshort = -1; bestcelllong = -1;
  for (icell=0; icell<NCELLHELIXSHORT;icell++){
    if ( cellcountshort[icell] > ncellmaxshort ) { 
      bestcellshort = icell; 
      ncellmaxshort = cellcountshort[icell]; 
    } 
  }
  for (icell=0; icell<NCELLHELIXLONG;icell++){
    if ( cellcountlong[icell] > ncellmaxlong ) { 
      bestcelllong = icell; 
      ncellmaxlong = cellcountlong[icell]; 
    } 
  }


  /*VB: aufgeteilt*/
  /* Loop over cell # arrays to uncover the traitors */
  for ( ifed=0; ifed<silraw->nfed; ifed++ ) {
    if ( silraw->fed[ifed].pipelen == NCELLHELIXSHORT ){
      if ( cell1[ifed] != bestcellshort && cell2[ifed] != bestcellshort )
	linkfail++;
      if ( cell1[ifed] != bestcellshort && ITR_silraw_evno > 10 ) {
	ITR_silraw_chipfail[2*ifed]++;
	if ( ITR_silraw_iol.cell_check_output.value == 2 )
	  aframe_mess(routine,GRR_SINFO,"Evt %d: Chip 0x%4x: Short cell # %3d differs from common cell # %3d",
		      ITR_silraw_evno, ITR_silraw_chipid[2*ifed], cell1[ifed], bestcellshort );
      }
      if ( cell2[ifed] != bestcellshort && ITR_silraw_evno > 10 ) {
	ITR_silraw_chipfail[2*ifed+1]++;
	if ( ITR_silraw_iol.cell_check_output.value == 2 )
	  aframe_mess(routine,GRR_SINFO, "Evt %d: Chip 0x%4x: Short cell # %3d differs from common cell # %3d",
		      ITR_silraw_evno, ITR_silraw_chipid[2*ifed+1], cell2[ifed], bestcellshort );
      }
    }
    else if (silraw->fed[ifed].pipelen == NCELLHELIXLONG){
      if ( cell1[ifed] != bestcelllong && cell2[ifed] != bestcelllong )
	linkfail++;
      if ( cell1[ifed] != bestcelllong && ITR_silraw_evno > 10 ) {
	ITR_silraw_chipfail[2*ifed]++;
	if ( ITR_silraw_iol.cell_check_output.value == 2 )
	  aframe_mess(routine,GRR_SINFO,"Evt %d: Chip 0x%4x: Long cell # %3d differs from common cell # %3d",
		      ITR_silraw_evno, ITR_silraw_chipid[2*ifed], cell1[ifed], bestcelllong );
      }
      if ( cell2[ifed] != bestcelllong && ITR_silraw_evno > 10 ) {
	ITR_silraw_chipfail[2*ifed+1]++;
	if ( ITR_silraw_iol.cell_check_output.value == 2 )
	  aframe_mess(routine,GRR_SINFO, "Evt %d: Chip 0x%4x: Long cell # %3d differs from common cell # %3d",
		      ITR_silraw_evno, ITR_silraw_chipid[2*ifed+1], cell2[ifed], bestcelllong );
      }
    }
    else {
      if(ITR_silraw_iol.debug_mode.value)
      aframe_mess(routine,GRR_SINFO,"pipelinelength %d for FED 0x%x not valid!",
		     silraw->fed[ifed].pipelen,silraw->fed[ifed].fedid);
    }  
  } /*end of for (..ifed..)-loop*/ 

 
  /*VB: */
  /* check most preferred cell number against cell number found by ITR_silraw */
  if ( bestcellshort != silraw->header.cellshort && bestcellshort != -1 ) {
    ITR_silraw_cellfail_tot_short++;
    aframe_mess(routine,GRR_SINFO, "Evt %d: Cell(short) # %d differs from ITR_silraw cell # %d",
	 	ITR_silraw_evno, bestcellshort, silraw->header.cellshort );
  }

  if ( bestcelllong != silraw->header.celllong && bestcelllong != -1 ) {
    ITR_silraw_cellfail_tot_long++;
    aframe_mess(routine,GRR_SINFO, "Evt %d: Cell(long) # %d differs from ITR_silraw cell # %d",
	 	ITR_silraw_evno, bestcelllong, silraw->header.celllong );
  }

  /* Count failure rates since event 10 */
  if ( ITR_silraw_evno >= 10 ) {
    ITR_silraw_linkfail_tot += linkfail;
    if ( linkfail > 0) 
      ITR_silraw_evfail_tot++;
  }

  /* Check, if Helixes have become bad :
   * Need 5 consecutive bad events => then dump out event, but only one */
  if ( linkfail > 70 && ITR_silraw_evno > 10 && !shit_happens ) {
    if ( shit_happens_event == ITR_silraw_evno-1 ) 
      shit_happens_count++;
    if ( shit_happens_count >= 5 ) {
      shit_happens = 1;
      mytime       = time(NULL);
      aframe_mess(routine,GRR_SINFO, "----------------------------------");
      aframe_mess(routine,GRR_SINFO, " GOT THE PROBLEM IN EVENT %d", ITR_silraw_evno );
      aframe_mess(routine,GRR_SINFO, asctime(localtime(&mytime)) );
      aframe_mess(routine,GRR_SINFO, "----------------------------------");
      for ( ifed=0; ifed<silraw->nfed; ifed++)
	aframe_mess(routine,GRR_SINFO, "link=%4d, fed=0x%x: cells are %3d %3d",
		    ifed, silraw->fed[ifed].fedid, cell1[ifed], cell2[ifed]);
    }
    shit_happens_event = ITR_silraw_evno;
  }
  else
    shit_happens_count = 0;

  /*VB: aufgeteilt...*/
  /* print out results */
  if ( ITR_silraw_evno < 15 && ITR_silraw_iol.cell_check_output.value ) {
    aframe_mess(routine,GRR_SINFO,
		"Best (short pipeline) cell # is %d, appears %d (%d) times.",
		bestcellshort, ncellmaxshort, 2*silraw->nfed );
    aframe_mess(routine,GRR_SINFO,
		"Best (long pipeline) cell # is %d, appears %d (%d) times.",
		bestcelllong, ncellmaxlong, 2*silraw->nfed );
    aframe_mess(routine,GRR_SINFO,
		"%d links failed (both helix cell #'s differ)",
		linkfail );
  }
  if ( linkfail > 0 && ( ITR_silraw_evno < 10000 || ITR_silraw_evno%100 == 0 )
                    && ITR_silraw_iol.cell_check_output.value      )
    aframe_mess(routine,GRR_SINFO, "Evt=%d: asynch links=%d",
		ITR_silraw_evno, linkfail );
  if ( linkfail > 70 && ( ITR_silraw_evno < 10000 || ITR_silraw_evno%100 == 0 )
                     && !ITR_silraw_iol.cell_check_output.value      )
    aframe_mess(routine,GRR_SINFO, "Evt=%d: asynch links=%d",
		ITR_silraw_evno, linkfail );
  if ( ITR_silraw_evno >= 100 && ITR_silraw_evno%400 == 1 )
    aframe_mess(routine,GRR_SINFO,
		"Evt %d: %d links in %d events out of sync since evt 10",
		ITR_silraw_evno, ITR_silraw_linkfail_tot, ITR_silraw_evfail_tot );

  /* increase event number */
  ITR_silraw_evno++;

  return GRR_OK;
}


/*-----------------------------------------------------------------------------
 * check an assumed offset for digitizings
 * =======================================
 * Given a candidate offset (for chips with id=0) this function checks
 * whether the first offset bytes are essentially constant, followed
 * by a jump in the next digitizing.
 * 
 * Author: Michael Schmelling / 2-Sep-1998
 *---------------------------------------------------------------------------*/
int ITR_silraw_check_offset(unsigned char *data, int offset)
{
  int   i, ifail=0;
  float val, s1, s2;
/*---------------------------------------------------------------------------*/

  for(i=0; i<offset; i++) {
    val = (float)*(data+i);
    s1 += val;
    s2 += val*val;
  }
  s1 = s1/(float)offset;
  s2 = s2/(float)offset - s1*s1;
  
  val = (float)*(data+offset) - s1;
  if(s2>2.|| val*val<10.) ifail++; 

  return ifail;

}

